#! /usr/bin/python
# encoding: latin-1

# ELF file parser

from elf_constants import *
import struct
import os
import bisect
import re

# The ELF header

## #define EI_NIDENT 16
## typedef struct {
## 	unsigned char	e_ident[EI_NIDENT];	/* File identification. */
## 	Elf32_Half	e_type;		/* File type. */
## 	Elf32_Half	e_machine;	/* Machine architecture. */
## 	Elf32_Word	e_version;	/* ELF format version. */
## 	Elf32_Addr	e_entry;	/* Entry point. */
## 	Elf32_Off	e_phoff;	/* Program header file offset. */
## 	Elf32_Off	e_shoff;	/* Section header file offset. */
## 	Elf32_Word	e_flags;	/* Architecture-specific flags. */
## 	Elf32_Half	e_ehsize;	/* Size of ELF header in bytes. */
## 	Elf32_Half	e_phentsize;	/* Size of program header entry. */
## 	Elf32_Half	e_phnum;	/* Number of program header entries. */
## 	Elf32_Half	e_shentsize;	/* Size of section header entry. */
## 	Elf32_Half	e_shnum;	/* Number of section header entries. */
## 	Elf32_Half	e_shstrndx;	/* Section name strings section. */
## } Elf32_Ehdr;

class ElfHeader:
    def __init__(self, fd):
        """
        fd is a file object pointing to the ELF header
        """
        oldfpointer = fd.tell() # save fd position
        fd.seek(0)
        header_format = "<16s2H5L6H"
        header_raw = fd.read(struct.calcsize(header_format))
        # now define class fields
        (self.e_ident,
        self.e_type,
        self.e_machine,
        self.e_version,
        self.e_entry,
        self.e_phoff,
        self.e_shoff,
        self.e_flags,
        self.e_ehsize,
        self.e_phentsize,
        self.e_phnum,
        self.e_shentsize,
        self.e_shnum,
        self.e_shstrndx) = struct.unpack(header_format, header_raw)
        # restore file pointer
        fd.seek(oldfpointer)
        # check ELF
        if self.e_ident[0:4] != ELFMAG:
            raise RuntimeError("Not an ELF file")


# The ELF section

## typedef struct {
## 	Elf32_Word	sh_name;	/* Section name (index into the
## 					   section header string table). */
## 	Elf32_Word	sh_type;	/* Section type. */
## 	Elf32_Word	sh_flags;	/* Section flags. */
## 	Elf32_Addr	sh_addr;	/* Address in memory image. */
## 	Elf32_Off	sh_offset;	/* Offset in file. */
## 	Elf32_Size	sh_size;	/* Size in bytes. */
## 	Elf32_Word	sh_link;	/* Index of a related section. */
## 	Elf32_Word	sh_info;	/* Depends on section type. */
## 	Elf32_Size	sh_addralign;	/* Alignment in bytes. */
## 	Elf32_Size	sh_entsize;	/* Size of each entry in section. */
## } Elf32_Shdr;

class ElfSection:
    def __init__(self, fd, index):
        """
        fd must be set to the position of the section to read
        fd will not be restored to its previous location
        index is the section index in the ELF file
        """
        section_format = "<10L"
        section_raw = fd.read(struct.calcsize(section_format))
        (self.sh_name,
        self.sh_type,
        self.sh_flags,
        self.sh_addr,
        self.sh_offset,
        self.sh_size,
        self.sh_link,
        self.sh_info,
        self.sh_addralign,
        self.sh_entsize) = struct.unpack(section_format, section_raw)
        # handle string tables
        if self.sh_type == SHT_STRTAB:
            # save file position
            oldposition = fd.tell()
            fd.seek(self.sh_offset)
            self.stringdata = fd.read(self.sh_size)
            fd.seek(oldposition)
        # To deal with RAM compression performed by armlink
        self.truesize = self.sh_size
        self.index = index
        self.name = ""

    def getstring(self, offset):
       """
       get the string for a string table, given the offset
       """
       if self.sh_type != SHT_STRTAB:
           return ''
       stringend = self.stringdata.find('\x00',offset+1)
       return self.stringdata[offset:stringend]

# The symbol table

## typedef struct {
## 	Elf32_Word	st_name;	/* String table index of name. */
## 	Elf32_Addr	st_value;	/* Symbol value. */
## 	Elf32_Size	st_size;	/* Size of associated object. */
## 	unsigned char	st_info;	/* Type and binding information. */
## 	unsigned char	st_other;	/* Reserved (not used). */
## 	Elf32_Half	st_shndx;	/* Section index of symbol. */
## } Elf32_Sym;
##
## /* Macros for accessing the fields of st_info. */
## #define ELF32_ST_BIND(info)		((info) >> 4)
## #define ELF32_ST_TYPE(info)		((info) & 0xf)

class ElfSymbolTable:
    def __init__(self, fd, symsh, strsh):
        """
        fd is a file object pointing to the ELF header (will be restored)
        symsh is a ElfSection holding the symbol table
        strsh is a ElfSection holding the string table
        """
        sym_format = "<3L2BH"
        oldposition = fd.tell()
        fd.seek(symsh.sh_offset)
        self.symbols = {}
        self.addresses = {}
        symcount = symsh.sh_size / symsh.sh_entsize
        symlength = struct.calcsize(sym_format) # should equal symsh.sh_entsize
        # for symbol name verification
        valid_sym_re = re.compile("^\w[a-zA-Z0-9_@]+$")
        for x in xrange(symcount):
            symraw = fd.read(symlength)
            (name, value, size, info, other, shndx) = struct.unpack(sym_format, symraw)
            # resolve the symbol name
            name = strsh.getstring(name)
            # check if the symbol is OK
            if valid_sym_re.search(name) is None:
                continue

            # get bind and type
            bind = info >> 4
            st_type = info & 0xf
            self.symbols[name] = { "value": value,
                    "size": size,
                    "bind" : bind,
                    "type" : st_type,
                    "shndx" : shndx }

            # for fast reverse lookup
            self.addresses[value] = name

        # restore file descriptor
        fd.seek(oldposition)
        # sort the address table
        self.locations = self.addresses.keys()
        self.locations.sort()

    def __getitem__(self, key):
        return self.symbols[key]["value"], self.symbols[key]["size"]

    def translate(self, address):
        """
        Converts an address into a symbol
        Returns a pair 'symbol name',symbol address
        """
        if self.addresses.has_key(address):
            return (self.addresses[address], 0)
        lPosition = bisect.bisect_left(self.locations, address)
        if lPosition == 0 or lPosition == len(self.locations):
            return (None, address)
        lNearest = self.locations[lPosition-1]
        if lNearest == 0:
            return (None, address)
        return (self.addresses[lNearest], address-lNearest)



# putting it all together

class ElfFile:
    def __init__(self, fd):
        """
        fd is a file object pointing to the ELF header
        """
        self.fd = fd
        # get the header
        self.header = ElfHeader(self.fd)
        # get the sections
        self.sections = {}
        # locate the string section first
        fd.seek(self.header.e_shoff)
        tmp_sections = []
        strsh = None
        symsh = None
        index = 0
        for x in xrange(self.header.e_shnum):
            section = ElfSection(self.fd,index)
            index += 1
            tmp_sections.append(section)
            if x == self.header.e_shstrndx:
                strsh = section
            if section.sh_type == SHT_SYMTAB:
                symsh = section
                symsh_str = section.sh_link
        # solve symbols
        if symsh is not None:
            self.symsh = ElfSymbolTable(fd, symsh, tmp_sections[symsh_str])
        # check if string section available
        if strsh is None:
            raise RuntimeError("No string section available")
        # now fill the sections hash
        for n in xrange(len(tmp_sections)):
            x = tmp_sections[n]
            name_offset = x.sh_name
            section_name = strsh.getstring(name_offset)
            x.name = section_name
            if section_name in self.sections:
                # check if compression issue exists, fix true size
                other_sect = self.sections[section_name]
                if other_sect.sh_type == SHT_PROGBITS and\
                   (other_sect.sh_flags & SHF_ALLOC):
                    other_sect.truesize = x.sh_addr - other_sect.sh_addr
                # rename
                section_name = "%s%d" % (section_name, n)

            self.sections[section_name] = x

    def get_section_address_and_data(self, section_name):
        """
        returns the raw contents of section "section_name"
        """
        (offset, size) = self.get_section_offset_and_size(section_name)
        self.fd.seek(offset)
        return (self.sections[section_name].sh_addr, self.fd.read(size))

    def get_section_offset_and_size(self, section_name):
        """
        returns (offset, size) for a given section
        """
        if not self.sections.has_key(section_name):
            raise RuntimeError("No such section %s" % section_name)
        offset = self.sections[section_name].sh_offset
        size = self.sections[section_name].sh_size
        return (offset, size)


    def sym2addr(self, symname):
        if symname in self.symsh.symbols:
            return self.symsh[symname]
        else:
            return None

    def addr2sym(self, address):
        return self.symsh.translate(address)


    def get_loadable_sections_list(self):
        """
        returns the list of sections which should be loaded in memory
        """
        list = []
        for section_name in self.sections.keys():
            section = self.sections[section_name]
            if (section.sh_flags & SHF_ALLOC) and ((section.sh_type == SHT_PROGBITS) or (section.sh_type == SHT_ARM_EXIDX)):
                list.append(section)

        namesList = []
        for section in sorted ( list, key=lambda section: section.index ):
            namesList.append ( section.name )
        return namesList


    def findsym(self, expr):
        """
        returns data symbols matching a reg exp
        """
        regexp = re.compile(expr, re.IGNORECASE)
        symbols = [x for x in self.symsh.symbols.keys() if regexp.search(x) is not None
                and self.symsh.symbols[x]["type"] == STT_OBJECT]
        symbols.sort()
        return symbols


    def is_addr_in_elf(self, addr):
        """
        check whether an address is in a section of this ELF file or not
        """
        for section_name in self.sections.keys():
            section = self.sections[section_name]
            base = section.sh_addr
            size = section.truesize
            if (section.sh_flags & SHF_ALLOC):
                # check if addr is within bounds
                if base <= addr and addr < base+size:
                    return True
        return False


    def get_load_size(self):
        """
        sum of the allocatable sections,
        must match the size of the firmware binary file
        """
        return sum([s.sh_size for s in self.sections.values() \
                if s.sh_type == SHT_PROGBITS and s.sh_flags & SHF_ALLOC])


def checkfile(filename):
    """
    Checks if the supplied filename is an ELF file
    """
    try:
        fd = file(filename, "rb")
        header = ElfHeader(fd)
        fd.close()
    except:
        return False
    return True

